# pdfplucker/core.py
import multiprocessing
from pathlib import Path
from pdfplucker.processor import process_batch, process_pdf, create_converter

try:
    multiprocessing.set_start_method("spawn", force=True)
except RuntimeError:
    # If the start method is already set, we can ignore the error
    pass

def pdfplucker(
    source: str | Path,
    output: str | Path ="./results",
    folder_separation: bool = False,
    images: str | Path | None = None,
    timeout: int = 600,
    workers: int = 4,
    force_ocr: bool = False,
    device: str = "AUTO",
    markdown: bool = False,
    amount: int = 0,
):
    """
    Process PDF files and extract information.
    
    Parameters:
    -----------
    source : str or Path
        Path to the PDF files (directory or just a single file)
    output : str or Path, default="./results"
        Path to save the processed information
    folder_separation : bool, default=False
        Create separate folders for each PDF
    images : str or Path, optional
        Path to save the extracted images (ignore if folder_separation is active)
    timeout : int, default=600
        Time limit in seconds for processing each PDF
    workers : int, default=4
        Number of parallel processes (threads)
    force_ocr : bool, default=False
        Force OCR even if the PDF is not scanned
    device : str, default="AUTO"
        Type of device for processing ("CUDA", "CPU" or "AUTO")
    markdown : bool, default=False
        Export the document in an additional markdown file
    amount : int, default=0
        Amount of files to process (0 for all)
    
    Returns:
    --------
    dict or bool
        If processing a batch, returns metrics dictionary
        If processing a single file, returns success status (bool)
    """
    
    source_path = Path(source)
    output_path = Path(output)
    image_path = Path(images) if images else None

    device = device.upper()

    print("=" * 50)
    print("\033[34mPDFPlucker CLI - Docling Wrapper\033[0m")
    print("=" * 50)
    print(f"Source path: {source}")
    print(f"Output path: {output}")
    print(f"Device type: {device}")
    print(f"Number of workers: {workers}")
    print(f"Force OCR: {'yes' if force_ocr else 'no'}")
    print(f"Timeout: {timeout} seconds")
    print(f"Save markdown: {'yes' if markdown else 'no'}")
    print(f"Folder separation: {'yes' if folder_separation else 'no'}")
    print(f"Images path: {images if images else 'not used'}")
    print(f"Amount of files to process: {amount if amount > 0 else 'all'}")
    print("=" * 50)
    print("Starting...")

    # Normalize paths for Windows compatibility
    source_path = Path(str(source_path).replace('\\', '/'))
    output_path = Path(str(output_path).replace('\\', '/'))
    if image_path:
        image_path = Path(str(image_path).replace('\\', '/'))
    
    if source_path.is_file():
        # Process single PDF
        doc_converter = create_converter(
            device=device.upper(),
            num_threads=workers,
            force_ocr=force_ocr,
        )
        
        if folder_separation:
            images_path = output_path / source_path.stem / "images"
        else:
            images_path = image_path or output_path / "images"
        
        images_path.mkdir(parents=True, exist_ok=True)
        
        sucess = process_pdf(
            source_path,
            output_path,
            images_path,
            doc_converter,
            folder_separation,
            markdown,
        )
        return sucess
    else:
        # Process batch of PDFs
        return process_batch(
            source=source_path,
            output=output_path,
            image_path=image_path,
            separate_folders=folder_separation,
            max_workers=workers,
            timeout=timeout,
            device=device.upper(),
            markdown=markdown,
            force_ocr=force_ocr,
            amount=amount if amount > 0 else None,
        )


# Version 0.3.9

import os
import gc
import json
import fitz
import time
import logging
import multiprocessing  # Changed back to multiprocessing
from pathlib import Path
from concurrent.futures import as_completed, TimeoutError
from pdfplucker.utils import format_result, link_subtitles, get_safe_executor, logger, Data
from docling.document_converter import DocumentConverter, PdfFormatOption
from docling.datamodel.base_models import InputFormat
from docling.datamodel.document import ConversionResult
from docling_core.types.doc import ImageRefMode
from docling.exceptions import ConversionError
from docling.datamodel.pipeline_options import (
    AcceleratorDevice,
    AcceleratorOptions,
    PdfPipelineOptions,
    EasyOcrOptions,
)

def convert_paths_to_strings(obj):
    """Recursively convert all Path objects to strings in a nested structure"""
    if isinstance(obj, Path):
        return str(obj)
    elif isinstance(obj, dict):
        return {k: convert_paths_to_strings(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [convert_paths_to_strings(item) for item in obj]
    else:
        return obj

def create_converter(device : str = 'CPU', num_threads : int = 4, ocr_lang: list = ['es', 'pt'], force_ocr: bool = False) -> DocumentConverter:
    ''' Create a DocumentConverter object with the pipeline options configured''' 
    pipeline_options = PdfPipelineOptions()
    pipeline_options.do_ocr = True
    pipeline_options.do_table_structure = True
    pipeline_options.table_structure_options.do_cell_matching = True 
    pipeline_options.ocr_options.lang = ocr_lang
    pipeline_options.generate_picture_images = True
    pipeline_options.do_picture_classification = True
    pipeline_options.do_formula_enrichment = True
    
    # Aggressive scaling for low memory mode
    pipeline_options.images_scale = 1
    
    if force_ocr:
        # Rapid OCR or Easy OCr
        ocr_options = EasyOcrOptions(force_full_page_ocr=True, lang=ocr_lang)
        pipeline_options.ocr_options = ocr_options
    
    # Device acceleration
    device_type = AcceleratorDevice.CUDA if device.upper() == 'CUDA' else AcceleratorDevice.CPU if device.upper() == 'CPU' else AcceleratorDevice.AUTO if device.upper() == 'AUTO' else AcceleratorDevice.AUTO
    pipeline_options.accelerator_options = AcceleratorOptions(num_threads=num_threads, device=device_type)
    
    converter = DocumentConverter(
        format_options={
            InputFormat.PDF: PdfFormatOption(pipeline_options=pipeline_options)
        }
    )

    return converter

# Reintroducing the worker function from 0.3.6
def _worker(source, output, image_path, doc_converter, separate_folders, markdown, queue):
    try:
        success = process_pdf(
            source,
            output,
            image_path,
            doc_converter,
            separate_folders,
            markdown
        )
        queue.put(success)
    except Exception as e:
        logger.error(f"Non treated error at _worker: {e}")
        queue.put(False)

def process_with_timeout(
    source: Path,
    output: Path,
    image_path: Path,
    doc_converter: DocumentConverter,
    separate_folders: bool = False,
    timeout: int = 600,
    markdown: bool = False
) -> bool:
    """ Process a single PDF with safety timeout """
    """ Returns True if successful, False otherwise """

    if multiprocessing.get_start_method() != 'spawn':
        multiprocessing.set_start_method('spawn', force=True)

    queue = multiprocessing.Queue()
    
    filename = os.path.basename(source)
    logger.info(f"Starting processing for '{filename}'")

    process = multiprocessing.Process(
        target=_worker,
        args=(source, output, image_path, doc_converter, separate_folders, markdown, queue) 
    )
    
    start_time = time.time()
    process.start()
    process.join(timeout)

    if process.is_alive():
        logger.error(f"Timeout after {timeout}s! Killing process for '{filename}'")
        process.terminate()
        process.join()
        return False
    
    try:
        if not queue.empty():
            result = queue.get()
            time_elapsed = time.time() - start_time
            if result:
                logger.info(f"Successfully processed '{filename}' in {time_elapsed:.2f}s")
            return result
        return False
    except Exception as e:
        logger.error(f"Error retrieving result: {e}")
        return False

def process_pdf(
        source: Path,
        output: Path,
        image_path: Path | None,
        doc_converter: DocumentConverter,
        separate_folders: bool | None = False,
        markdown: bool = False,
) -> bool:
    """Function to process a single PDF file utilizing Docling"""

    conv = None
    filename = Path(os.path.basename(source))
    base_filename = Path(os.path.splitext(filename)[0])
    logger.debug(f"Starting PDF processing for {filename}")

    try:
        if separate_folders:
            specific_folder = Path(os.path.join(output, base_filename))
            result = Path(os.path.join(specific_folder, f"{base_filename}.json"))
            image_folder = Path(os.path.join(specific_folder, "images"))
            os.makedirs(specific_folder, exist_ok=True)
            os.makedirs(image_folder, exist_ok=True)
        else:
            result = Path(os.path.join(output, f"{base_filename}.json"))
            image_folder = Path(image_path)
    
        data: Data = {
            "metadata": {},
            "sections" : [],
            "images": [],
            "tables": [],
            "subtitles" : []
        }

        # Use PyMuPDF (fitz) for extracting metadata - lightweight operation
        with fitz.open(source) as doc:
            # Check number of pages for large documents
            num_pages = len(doc)
            if num_pages > 100:
                logger.warning(f"Large document detected: {filename} has {num_pages} pages")
            data["metadata"] = doc.metadata
            data["metadata"]["filename"] = filename
            data["metadata"]["pages"] = num_pages

        conv: ConversionResult = doc_converter.convert(str(source))
        format_result(conv, data, base_filename, image_folder)
        link_subtitles(data)

        # Save Markdown if asked - after clearing conversion memory
        if markdown:
            try:
                md_filename = Path(os.path.join(specific_folder, f"{base_filename}.md"))
                conv.document.save_as_markdown(md_filename, image_mode=ImageRefMode.EMBEDDED)
            except Exception as md_error:
                logger.error(f"Error saving markdown: {md_error}")

        # transform Paths into strings
        data = convert_paths_to_strings(data)

        with open(result, 'w', encoding='utf-8') as f:
            json.dump(data, f, ensure_ascii=False, indent=4)

        return True

    except MemoryError:
        logger.error(f"Out of memory while converting '{filename}'")
        gc.collect()
        return False
    except (fitz.FileDataError, fitz.EmptyFileError) as e:
        logger.error(f"Failed to process '{filename}': {e}")
        return False
    except IOError as e:      
        logger.error(f"I/O error while processing '{filename}': {e}")
        return False
    except Exception as e:    
        import traceback
        logger.error(f"Error processing '{filename}': {str(e)}\n{traceback.format_exc()}")
        return False
    except ConversionError as e:
        logger.error(f"Conversion error for '{filename}': {e}")
        return False
    
    finally:
        gc.collect()  # Aggressive garbage collection
        try:
            del conv
            del data
        except Exception as e:
            logger.error(f"Error during cleanup: {e}")

def process_batch(
    source: Path,
    output: Path,
    image_path: Path | None,
    separate_folders: bool = False,
    max_workers: int = 4,
    timeout: int = 600,
    device: str = 'AUTO',
    markdown: bool = False,
    force_ocr: bool = False,
    amount: int = None,
) -> dict:
    """ Process a batch of PDFs at a time in parallel """

    if not separate_folders and image_path is None:
        image_path = Path(f"{output}/images")
    
    # Create output directories
    os.makedirs(output, exist_ok=True)
    if not separate_folders:
        os.makedirs(image_path, exist_ok=True)
    
    # create the doc_converter
    doc_converter = create_converter(device=device, num_threads=max_workers, force_ocr=force_ocr)

    pdf_files = []

    if source.is_dir():
        for file in os.listdir(source):
            if file.lower().endswith('.pdf'):
                file_path = Path(os.path.join(source, file))
                pdf_files.append(file_path)
    pdf_files = pdf_files[:amount] if amount and amount > 0 else pdf_files

    total = len(pdf_files)
    logger.info(f"Total documents to process: {total}")

    # create the metrics
    metrics = {
        'initial_time': time.time(),
        'elapsed_time': 0,
        'total_docs': total,
        'processed_docs': 0,
        'failed_docs': 0,
        'timeout_docs': 0,
        'success_rate': 0,
        'memory_peak': 0,
        'fails': []
    }

    # Switch back to ProcessPoolExecutor
    with get_safe_executor(max_workers=max_workers) as executor:
        futures = {}
        for pdf_file in pdf_files:
            future = executor.submit(
                process_with_timeout,
                pdf_file,
                output,
                image_path if image_path else None,
                doc_converter,
                separate_folders,
                timeout,
                markdown
            )
            futures[future] = pdf_file

        for future in as_completed(futures):
            pdf_file = futures[future]
            metrics['processed_docs'] += 1
            
            try:
                success = future.result()
                if not success:
                    metrics['failed_docs'] += 1
                    metrics['fails'].append({
                        'file': str(pdf_file),
                        'error': 'Processing error'
                    })
            except TimeoutError:
                logger.error(f"Timeout reached for '{os.path.basename(str(pdf_file))}'")
                metrics['timeout_docs'] += 1
                metrics['fails'].append({
                    'file': str(pdf_file),
                    'error': "Timeout reached"
                })
            except Exception as e:
                logger.error(f"Processing error for '{os.path.basename(str(pdf_file))}': {e}")
                metrics['failed_docs'] += 1
                metrics['fails'].append({
                    'file': str(pdf_file),
                    'error': str(e)
                })

            gc.collect()

            if metrics['processed_docs'] % 5 == 0:
                # Save intermediate metrics every 5 files
                _update_metrics(metrics, output)

    # Finalize metrics
    del doc_converter
    gc.collect()
    _update_metrics(metrics, output, final=True)
    logger.info(f"Processing concluded, sucess rate: {metrics['success_rate']:.1f}%, total time: {metrics['elapsed_time']:.1f}s")

    return metrics

def _update_metrics(metrics: dict, output_dir: str, final: bool = False) -> None:
    """Atualiza e salva as métricas de processamento"""
    metrics['elapsed_time'] = time.time() - metrics['initial_time']

    # Convert Paths to string
    metrics = convert_paths_to_strings(metrics)

    processed = metrics['processed_docs']
    if processed > 0:
        metrics['success_rate'] = ((processed - metrics['failed_docs'] - metrics['timeout_docs']) / processed) * 100
    
    filename = 'final_metrics.json' if final else 'intermediate_metrics.json'
    with open(os.path.join(output_dir, filename), 'w') as f:
        json.dump(metrics, f, indent=2)


"""
pdfplucker - a powerful wrapper for the Docling library
"""
# __init__.py

__version__ = "0.3.9"
__author__ = "Rafael Ghiorzi"
__email__ = "rafael.ghiorzi@gmail.com"

from pdfplucker.processor import process_batch, process_pdf, create_converter, process_with_timeout, _update_metrics
from pdfplucker.utils import format_result, link_subtitles, setup_logging, get_safe_executor
from pdfplucker.core import pdfplucker

__all__ = [
    "setup_logging",
    "get_safe_executor",
    "process_with_timeout",
    "_update_metrics",
    "process_batch",
    "process_pdf",
    "create_converter",
    "format_result",
    "link_subtitles",
    "pdfplucker",
]


# utils.py
from concurrent.futures import ProcessPoolExecutor
import multiprocessing
from docling.datamodel.document import ConversionResult
from docling_core.types.doc import (
    PictureItem,
    TableItem,
    TextItem,
    DocItemLabel,
)
from typing import TypedDict, List, Dict, Any
import warnings
import logging
import os

class Data(TypedDict):
    metadata: Dict[str, Any]
    sections: List[Dict[str, Any]]
    images: List[Dict[str, Any]]
    tables: List[Dict[str, Any]]
    subtitles: List[Dict[str, Any]]

def format_result(conv: ConversionResult, data: Data, filename: str, image_path: str) -> None:
    try:
        ''' Uses the docling document to format a readable JSON result '''

        collecting = None
        counter = 0

        for idx, (item, _) in enumerate(conv.document.iterate_items()):
            if isinstance(item, TextItem):
                if item.label == DocItemLabel.SECTION_HEADER:
                    if collecting is not None:
                        data['sections'].append(collecting)
                    collecting = {'title': item.text, 'text': ''}
                elif item.label == DocItemLabel.FORMULA:
                    if collecting is not None:
                        collecting['text'] += '\n' + "Equation:" + item.text if collecting['text'] else item.text
                else:
                    if collecting is not None:
                        collecting['text'] += '\n' + item.text if collecting['text'] else item.text
            elif isinstance(item, TableItem):
                table = item.export_to_dataframe()
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", UserWarning)
                    table_dict = table.to_dict()
                data['tables'].append({
                    'self_ref' : item.self_ref,
                    'subtitle' : '',
                    'table' : table_dict
                })
            elif isinstance(item, PictureItem):
                classification = None
                if item.annotations:
                    for annotation in item.annotations:
                        if annotation.kind == 'classification':
                            # Find the classification with the highest confidence
                            best_class = max(
                                annotation.predicted_classes,
                                key=lambda cls: cls.confidence
                            )
                            classification = {
                                'class_name': best_class.class_name,
                                'confidence': best_class.confidence
                            }
                            break

                image_filename = (image_path  / f"{filename}_{counter}.png")
                with image_filename.open('wb') as f:
                    item.get_image(conv.document).save(f, "PNG")
                data['images'].append({
                    'ref': f"{filename}_{counter}.png",
                    'self_ref' : item.self_ref,
                    'classification' : classification,
                    'subtitle' : ''
                })
                counter += 1

        # Collecting the results after all iterations
        if collecting is not None:
            data['sections'].append(collecting)
        for text in conv.document.texts:
            if text.label == 'caption':
                data['subtitles'].append({
                    'ref' : text.parent.cref,
                    'text' : text.text
                })
    except Exception as e:
        print(f"\033[33mError formatting result: {e}\033[0m")

def link_subtitles(data: Data) -> None:
    ''' 
    Try associating subtitles with images or tables
    based on refs saved on the Docling document
    '''
    try:
        images = data.get('images', [])
        tables = data.get('tables', [])
        subtitles = data.get('subtitles', [])

        # iterate over all images and all subtitles
        for img in images:
            self_ref = img.get('self_ref')
            if self_ref is None:
                continue
            
            # For each image, try to find a ref equals to the image self ref
            for sub in subtitles[:]:
                ref = sub.get('ref')
                if self_ref == ref:
                    img['subtitle'] = sub.get('text', '')
                    img.pop('item', None)
                    # remove the subtitle to be sure
                    subtitles.remove(sub)
                    break 

        # iterate over all tables and the remaining subtitles
        for tab in tables:
            self_ref = tab.get('self_ref')
            if self_ref is None:
                continue

            # For each table, try to find a ref equals to the table self ref
            for sub in subtitles[:]:
                ref = sub.get('ref')
                if self_ref == ref:
                    tab['subtitle'] = sub.get('text', '')
                    subtitles.remove(sub)
                    break

        for sub in subtitles:
            data['images'].append({
                'ref': sub.get('ref'),
                'self_ref' : None,
                'subtitle' : sub.get('text', '')
            })

        data.pop('subtitles')
    except Exception as e:
        print(f"\033[33mError linking subtitles: {e}\033[0m]")
    finally:
        if 'subtitles' in locals():
            del subtitles
        if 'images' in locals():
            del images
        if 'tables' in locals():
            del tables

def get_safe_executor(max_workers=None):
    try:
        multiprocessing.set_start_method('spawn', force=True)
    except RuntimeError:
        # Method might already be set
        pass
        
    return ProcessPoolExecutor(max_workers=max_workers)

class ColorfulFormatter(logging.Formatter):
    COLOURS = {
        'DEBUG': '\033[94m',  # Blue
        'INFO': '\033[92m',   # Green
        'WARNING': '\033[93m',  # Yellow
        'ERROR': '\033[91m',  # Red
        'CRITICAL': '\033[95m',  # Magenta
    }
    RESET = '\033[0m'  # Reset to default color

    def format(self, record):
        log_color = self.COLOURS.get(record.levelname, self.RESET)
        record.levelname = f"{log_color}{record.levelname}{self.RESET}"
   
        return super().format(record)

def setup_logging(level=logging.INFO):
    # Create formatters
    detailed_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    simple_formatter = ColorfulFormatter('%(asctime)s - %(levelname)s - %(message)s')
    
    # Create file handlers
    log_dir = os.path.join(os.getcwd(), 'logs')
    os.makedirs(log_dir, exist_ok=True)
    
    pdfplucker_file = logging.FileHandler(os.path.join(log_dir, 'pdfplucker.log'))
    pdfplucker_file.setFormatter(detailed_formatter)
    
    # Create console handler
    console = logging.StreamHandler()
    console.setFormatter(simple_formatter)
    
    # Create a third-party log file for docling
    third_party_file = logging.FileHandler(os.path.join(log_dir, 'dependencies.log'))
    third_party_file.setFormatter(detailed_formatter)
    third_party_file.setLevel(logging.DEBUG)  # All levels
    
    # Configure the root logger (for third-party libraries)
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)  # All levels
    
    # Remove any existing handlers to avoid duplicates
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)
    
    root_logger.addHandler(third_party_file)
    
    # Configure your package's logger
    pdfplucker_logger = logging.getLogger('pdfplucker')
    pdfplucker_logger.setLevel(level)
    
    # Remove any existing handlers to avoid duplicates
    for handler in pdfplucker_logger.handlers[:]:
        pdfplucker_logger.removeHandler(handler)
    
    pdfplucker_logger.addHandler(pdfplucker_file)
    pdfplucker_logger.addHandler(console)
    
    # Make sure your logger doesn't propagate to root
    pdfplucker_logger.propagate = False
    
    return pdfplucker_logger

logger = setup_logging()
